import json
import random
import unittest
from unittest.mock import MagicMock, patch

import numpy as np

from skllm.models.gpt_zero_shot_clf import (
    MultiLabelZeroShotGPTClassifier,
    ZeroShotGPTClassifier,
)


def _get_ret(label):
    return {"choices": [{"message": {"content": json.dumps({"label": label})}}]}


class TestZeroShotGPTClassifier(unittest.TestCase):
    def get_mock_clf_model(self):
        clf = ZeroShotGPTClassifier(
            openai_key="mock_key", openai_org="mock_org"
        )  # Mock keys
        X = np.array(["text1", "text2", "text3"])
        y = np.array(["class1", "class2", "class1"])
        clf.fit(X, y)

        self.assertEqual(set(clf.classes_), set(["class1", "class2"]))
        self.assertEqual(clf.probabilities_, [2 / 3, 1 / 3])
        return clf

    @patch(
        "skllm.models.gpt_zero_shot_clf.get_chat_completion", return_value=MagicMock()
    )
    def test_fit_predict(self, mock_get_chat_completion):
        clf = self.get_mock_clf_model()
        mock_get_chat_completion.return_value.choices[0].message = {
            "content": json.dumps({"label": "class1"})
        }
        predictions = clf.predict(["text1", "text2", "text3"])

        self.assertEqual(predictions, ["class1", "class1", "class1"])

    @patch(
        "skllm.models.gpt_zero_shot_clf.get_chat_completion", return_value=MagicMock()
    )
    def test_fit_predict_unknown_label_set_default(self, mock_get_chat_completion):
        clf = self.get_mock_clf_model()

        # Random choice
        mock_get_chat_completion.return_value.choices[0].message = {
            "content": json.dumps({"label": "new_class"})
        }
        clf.probabilities_ = [0, 1]
        predictions = clf.predict(["text1", "text2", "text3"])
        random.seed(42)
        np.random.seed(42)

        self.assertEqual(predictions, ["class2", "class2", "class2"])  # Random choice

        clf.default_label = "default_class"
        predictions = clf.predict(["text1", "text2", "text3"])
        self.assertEqual(
            predictions, ["default_class", "default_class", "default_class"]
        )

        clf.default_label = None
        predictions = clf.predict(["text1", "text2", "text3"])
        self.assertEqual(predictions, [None, None, None])


class TestMultiLabelZeroShotGPTClassifier(unittest.TestCase):
    def get_mock_clf_model(self):
        clf = MultiLabelZeroShotGPTClassifier(
            openai_key="mock_key", openai_org="mock_org", default_label="Random"
        )  # Mock keys
        X = np.array(["text1", "text2", "text3"])
        y = [
            ["class1", "class2"],
            ["class1", "class2"],
            ["class1", "class2"],
        ]  # Adjusted y to ensure [0.5, 0.5] probability
        clf.fit(X, y)

        self.assertEqual(set(clf.classes_), set(["class1", "class2"]))
        self.assertEqual(clf.probabilities_, [0.5, 0.5])
        return clf

    @patch(
        "skllm.models.gpt_zero_shot_clf.get_chat_completion", return_value=MagicMock()
    )
    def test_fit_predict(self, mock_get_chat_completion):
        clf = self.get_mock_clf_model()
        mock_get_chat_completion.return_value = _get_ret(["class1", "class2"])
        predictions = clf.predict(["text1", "text2", "text3"])
        self.assertEqual(
            predictions,
            [["class1", "class2"], ["class1", "class2"], ["class1", "class2"]],
        )

    @patch(
        "skllm.models.gpt_zero_shot_clf.get_chat_completion", return_value=MagicMock()
    )
    def test_fit_predict_unknown_label_set_default(self, mock_get_chat_completion):
        clf = self.get_mock_clf_model()
        mock_get_chat_completion.return_value.choices[0].message = {
            "content": json.dumps({"label": "new_class"})
        }

        clf.probabilities_ = [0.0, 1.0]
        predictions = clf.predict(["text1", "text2", "text3"])
        random.seed(42)
        np.random.seed(42)
        self.assertEqual(
            predictions, [["class2"], ["class2"], ["class2"]]
        )  # Random choice

        clf.default_label = ["default_class"]
        predictions = clf.predict(["text1", "text2", "text3"])
        self.assertEqual(
            predictions, [["default_class"], ["default_class"], ["default_class"]]
        )

        clf.default_label = None
        predictions = clf.predict(["text1", "text2", "text3"])
        self.assertEqual(predictions, [None, None, None])


if __name__ == "__main__":
    unittest.main()
